#include "invertSparseCov.h"

/* Implementation of routines (begin) --------------------------------------- */

/* -------------------------------------------------------------------------- */
/* Implementation of class "ConstSparseSqMatrix" (constructor)                */
/* -------------------------------------------------------------------------- */
ConstSparseSqMatrix::ConstSparseSqMatrix(long long* offsets, double* elements,
	long long* numElements, int* numCol) {
	// Fill the direct fields
	this->numRow = *numCol;
	this->numCol = *numCol;
	this->numElements = *numElements;
	this->elementsByCol = elements;
	this->offsets = offsets;
	// Specify lengths of columns and rows and column offsets
	this->colLength = new int[this->numCol];
	this->colOffset = new long long[this->numCol];
	for (int i = 0; i < this->numCol; i++) {
		colLength[i] = 0;
		colOffset[i] = -1;
	}
	this->rowLength = new int[this->numRow];
	for (int i = 0; i < this->numRow; i++) { rowLength[i] = 0; }
	this->colIndex = new int[this->numElements];
	this->rowIndex = new int[this->numElements];
	int curICol = -1;
	for (long long i = 0; i < this->numElements; i++) {
		// Calculate row and column indices
		int iCol = (offsets[i] - 1) / this->numRow;
		int iRow = (offsets[i] - 1) % this->numRow;
		this->colIndex[i] = iCol;
		this->rowIndex[i] = iRow;
		// Update row length, column length, column offset
		colLength[iCol]++;
		rowLength[iRow]++;
		if (iCol > curICol) {
			colOffset[iCol] = i;
			curICol = iCol;
		}
	}
	// Index the matrix
	this->indexInCol = new int*[this->numCol]; // index inside of each column
	this->offsetInCol = new int*[this->numCol]; // offset inside of each column
	for (int i = 0; i < this->numCol; i++) {
		indexInCol[i] = new int[this->colLength[i]];
		offsetInCol[i] = new int[this->colLength[i]];
	}
	this->indexInRow = new int*[this->numRow]; // index inside of each row
	this->offsetInRow = new int*[this->numRow]; // offset inside of each row
	for (int i = 0; i < this->numRow; i++) {
		indexInRow[i] = new int[this->rowLength[i]];
		offsetInRow[i] = new int[this->rowLength[i]];
	}
	int* curColLength = new int[this->numCol]; // cur length in each column
	for (int i = 0; i < this->numCol; i++) { curColLength[i] = 0; }
	int* curRowLength = new int[this->numRow]; // cur length in each row
	for (int i = 0; i < this->numRow; i++) { curRowLength[i] = 0; }
	for (long long i = 0; i < this->numElements; i++) {
		// Index
		indexInCol[this->colIndex[i]][curColLength[this->colIndex[i]]] = this->rowIndex[i];
		offsetInCol[this->colIndex[i]][curColLength[this->colIndex[i]]] = i;
		curColLength[this->colIndex[i]]++;
		indexInRow[this->rowIndex[i]][curRowLength[this->rowIndex[i]]] = this->colIndex[i];
		offsetInRow[this->rowIndex[i]][curRowLength[this->rowIndex[i]]] = i;
		curRowLength[this->rowIndex[i]]++;
	}
	// Release memory
	delete[] curColLength;
	delete[] curRowLength;
}

/* -------------------------------------------------------------------------- */
/* Find index of 'val' in the ordered array 'start' of length 'length'        */
/* -------------------------------------------------------------------------- */
int FindIndex(int val, int* start, int length) {
	int left = 0;
	int right = length;
	while (left < right) {
		int middle = (left + right) / 2;
		if (start[middle] < val) {
			left = middle;
		}
		else {
			if (start[middle] > val) {
				right = middle;
			}
			else {
				return middle;
			}
		}
	}
	return left;
}

/* -------------------------------------------------------------------------- */
/* Copy vector 'input' into vector 'output' unique and ordered                */
/* -------------------------------------------------------------------------- */
int OrderVector(vector<int> &input, vector<int> &output) {
	output.resize(0);
	sort(input.begin(), input.end());
	int curVal = input[0];
	output.push_back(curVal);
	for (int i = 1; i < input.size(); i++) {
		if (input[i] == curVal) { continue; }
		curVal = input[i];
		output.push_back(curVal);
	}
	return 0;
}

/* -------------------------------------------------------------------------- */
/* Determine elements of the covariance matrix to be calculated.              */
/* The routine is fast due to the use of "binaryMatrixLower", but             */
/* requires ceil(n * (n + 1) / 2 / 8) bytes of physical memory                */
/* -------------------------------------------------------------------------- */
int DetermineElements(ConstSparseSqMatrix &m, long long* offsetsNeeded,
	long long numElementsNeeded, long long* offsets, long long* numOffsets) {
	// Create the indicator matrix for 'S'
	binaryMatrixLower<unsigned long long> b =
		binaryMatrixLower<unsigned long long>(m.numCol);
	vector<long long> givenOffsets(0);
	for (long long i = 0; i < m.numElements; i++) {
		b.setIfNotSet(m.rowIndex[i], m.colIndex[i]);
		givenOffsets.push_back(m.offsets[i] - 1);
	}
	// Create vector of columns
	vector<vector<int>> cols(m.numCol);
	for (int i = 0; i < m.numCol; i++) {
		cols[i] = vector<int>(m.indexInCol[i], m.indexInCol[i] + m.colLength[i]);
	}
	// Add needed elements
	for (long long i = 0; i < numElementsNeeded; i++) {
		// Calculate row and column indices
		int iCol = (offsetsNeeded[i] - 1) / m.numRow;
		int iRow = (offsetsNeeded[i] - 1) % m.numRow;
		// Add to the coresponding column if new
		if (b.setIfNotSet(iRow, iCol)) {
			cols[iCol].push_back(iRow);
		}
	}
	// Determine needed indices column-by-column
	long long totSize = 0;
	for (int i = 0; i < m.numCol; i++) {
		totSize += cols[i].size();
		sort(cols[i].begin(), cols[i].end());
		for (int j = 0; j < cols[i].size(); j++) {
			if (cols[i][j] == i) { continue; }
			for (int k = 0; k < m.colLength[i]; k++) {
				if (m.indexInCol[i][k] == i) { continue; }
				if (m.indexInCol[i][k] <= cols[i][j]) {
					if (b.setIfNotSet(cols[i][j], m.indexInCol[i][k])) {
						cols[m.indexInCol[i][k]].push_back(cols[i][j]);
					}
				}
				if (m.indexInCol[i][k] > cols[i][j]) {
					if (b.setIfNotSet(m.indexInCol[i][k], cols[i][j])) {
						cols[cols[i][j]].push_back(m.indexInCol[i][k]);
					}
				}
			}
		}
	}
	// Collect the results
	*numOffsets = totSize;
	long long index = 0;
	for (int i = 0; i < m.numCol; i++) {
		for (int j = 0; j < cols[i].size(); j++) {
			offsets[index++] = (long long)i * (long long)m.numRow + cols[i][j] + 1;
		}
	}
	return m.numCol;
}

/* -------------------------------------------------------------------------- */
/* Determine elements of the covariance matrix to be calculated.              */
/* The routine allocates memory dynamically for nonzero elements only, which  */
/* makes it slower than when using "binaryMatrixLower"                        */
/* -------------------------------------------------------------------------- */
int DetermineElementsThrifty(ConstSparseSqMatrix &m, long long* offsetsNeeded,
	long long numElementsNeeded, long long* offsets, long long* numOffsets) {
	// Create the indicator matrix for 'S'
	vector<long long> givenOffsets(0);
	for (long long i = 0; i < m.numElements; i++) {
		givenOffsets.push_back(m.offsets[i] - 1);
	}
	// Create vector of columns
	vector<vector<int>> colsAll(m.numCol);
	for (int i = 0; i < m.numCol; i++) {
		colsAll[i] = vector<int>(m.indexInCol[i], m.indexInCol[i] + m.colLength[i]);
	}
	vector<vector<int>> cols(m.numCol);
	// Add needed elements
	for (long long i = 0; i < numElementsNeeded; i++) {
		// Calculate row and column indices
		int iCol = (offsetsNeeded[i] - 1) / m.numRow;
		int iRow = (offsetsNeeded[i] - 1) % m.numRow;
		colsAll[iCol].push_back(iRow);
		// Add to the coresponding column if new
	}
	// Determine needed indices column-by-column
	long long totSize = 0;
	for (int i = 0; i < m.numCol; i++) {
		OrderVector(colsAll[i], cols[i]);
		totSize += cols[i].size();
		//sort(cols[i].begin(), cols[i].end());
		for (int j = 0; j < cols[i].size(); j++) {
			if (cols[i][j] == i) { continue; }
			for (int k = 0; k < m.colLength[i]; k++) {
				if (m.indexInCol[i][k] == i) { continue; }
				if (m.indexInCol[i][k] <= cols[i][j]) {
					colsAll[m.indexInCol[i][k]].push_back(cols[i][j]);
				}
				if (m.indexInCol[i][k] > cols[i][j]) {
					colsAll[cols[i][j]].push_back(m.indexInCol[i][k]);
				}
			}
		}
	}
	// Collect the results
	*numOffsets = totSize;
	long long index = 0;
	for (int i = 0; i < m.numCol; i++) {
		for (int j = 0; j < cols[i].size(); j++) {
			offsets[index++] = (long long)i * (long long)m.numRow + cols[i][j] + 1;
		}
	}
	return m.numCol;
}

/* -------------------------------------------------------------------------- */
/* Calculate all the nonzero entries of the matrix                            */
/* -------------------------------------------------------------------------- */
int FillCov(ConstSparseSqMatrix &lowerCholesky, ConstSparseSqMatrix &lowerCov) {
	// Calculate the elements of 'lowerCov'
	for (int i = lowerCov.numCol - 1; i >= 0; i--) {
		for (int j = lowerCov.colLength[i] - 1; j >= 0; j--) {
			// Elementwise application of the recursion formula by Takahashi, Fagan, Chen (1973)
			int curRow = lowerCov.indexInCol[i][j];
			long long offsetInCurCol = 1;
			double acc = 0;
			if (curRow == i) {
				acc += 1. / (lowerCholesky.elementsByCol[lowerCholesky.colOffset[i]]
					* lowerCholesky.elementsByCol[lowerCholesky.colOffset[i]]);
			}
			else {
				int start = FindIndex(i, lowerCov.indexInRow[curRow], lowerCov.rowLength[curRow]) + 1;
				for (int k = start; k < lowerCov.rowLength[curRow]; k++) {
					if (lowerCov.indexInRow[curRow][k] < lowerCholesky.indexInCol[i][offsetInCurCol]) { continue; }
					acc -= 1. / lowerCholesky.elementsByCol[lowerCholesky.colOffset[i]]
						* (lowerCholesky.elementsByCol[lowerCholesky.colOffset[i] + offsetInCurCol++]
							* lowerCov.elementsByCol[(long long)lowerCov.offsetInRow[curRow][k]]);
					if (offsetInCurCol == lowerCholesky.colLength[i]) { break; }
				}
			}
			if (offsetInCurCol < lowerCholesky.colLength[i]) {
				for (int k = 1; k < lowerCov.colLength[curRow]; k++) {
					if (lowerCov.indexInCol[curRow][k] < lowerCholesky.indexInCol[i][offsetInCurCol]) { continue; }
					acc -= 1. / lowerCholesky.elementsByCol[lowerCholesky.colOffset[i]]
						* (lowerCholesky.elementsByCol[lowerCholesky.colOffset[i] + offsetInCurCol++]
							* lowerCov.elementsByCol[lowerCov.offsetInCol[curRow][k]]);
					if (offsetInCurCol == lowerCholesky.colLength[i]) { break; }
				}
			}
			lowerCov.elementsByCol[lowerCov.colOffset[i] + (long long)j] = acc;
		}
	}
	return lowerCov.numCol;
}

/* -------------------------------------------------------------------------- */
/* Check for mistakes in the input arguments                                  */
/* -------------------------------------------------------------------------- */
void errorCheck(int nlhs, int nrhs, const mxArray *prhs[]) {
	if (nrhs > 3) {
		mexErrMsgIdAndTxt("MATLAB:indexInvertSparse:rhs", "No more than tree input arguments allowed.");
	}
	if (nlhs > 1) {
		mexErrMsgIdAndTxt("MATLAB:indexInvertSparse:lhs", "One output argument is expected.");
	}
	if (mxGetN(prhs[0]) != mxGetM(prhs[0]) || mxGetN(prhs[0]) == 0 || mxGetM(prhs[0]) == 0) {
		mexErrMsgIdAndTxt("MATLAB:indexInvertSparse:rhs", "First input argument shoul be a nonempty sparse matrix.");
	}
	if (nrhs == 2) {
		if ((mxGetN(prhs[1]) != 2) &&
			(!mxIsDouble(prhs[1]) ||
			mxIsComplex(prhs[1]) ||
			mxGetNumberOfElements(prhs[1]) != 1)) {
			mexErrMsgIdAndTxt("MATLAB:indexInvertSparse:rhs", "Second input agrument must be either a two-column (possibly empty) matrix or a scalar.");
		}
	}
	if (nrhs == 3) {
		if (mxGetN(prhs[1]) != 2) {
			mexErrMsgIdAndTxt("MATLAB:indexInvertSparse:rhs", "Second input agrument must be a two-column (possibly empty) matrix.");
		}
		if (!mxIsDouble(prhs[2]) ||
			mxIsComplex(prhs[2]) ||
			mxGetNumberOfElements(prhs[2]) != 1) {
			mexErrMsgIdAndTxt("MATLAB:indexInvertSparse:rhs", "Last input agrument must be a scalar.");
		}
	}
}

/* Implementation of routines (end) ----------------------------------------- */

/* -------------------------------------------------------------------------- */
/* The MATLAB gateway function
/* -------------------------------------------------------------------------- */
void mexFunction(int nlhs, mxArray *plhs[],
	int nrhs, const mxArray *prhs[])
{
	// Check for input errors
	errorCheck(nlhs, nrhs, prhs);
	// Read input data
	mwSize n = mxGetN(prhs[0]);
	double* inPr = mxGetPr(prhs[0]);
	mwIndex* inIr = mxGetIr(prhs[0]);
	mwIndex* inJc = mxGetJc(prhs[0]);
	// Format input data
	long long* inOffsets = new long long[inJc[n]];
	// Determine the first (nonempty) column index
	mwIndex thisColIndex = 0;
	mwIndex nextColIndex = thisColIndex + 1;
	// Loop through columns
	mwIndex curIndex = 0;
	while (thisColIndex < n) {
		nextColIndex = thisColIndex + 1;
		// Pass empty columns
		while ((inJc[nextColIndex] == inJc[thisColIndex]) && (nextColIndex < n + 1)) {
			nextColIndex++;
		}
		thisColIndex = nextColIndex - 1;
		mwIndex thisRowIndex = inJc[thisColIndex];
		mwIndex lastRowIndex = inJc[nextColIndex];
		// Loop through rows
		while (thisRowIndex < lastRowIndex) {
			inOffsets[curIndex++] = thisColIndex * n + inIr[thisRowIndex++] + 1;
		}
		thisColIndex++;
		nextColIndex = thisColIndex + 1;
	}
	long long longNnz = inJc[n];
	int intN = n;
	ConstSparseSqMatrix lowerCholesky = ConstSparseSqMatrix(inOffsets, inPr, &longNnz, &intN);
	mwSize numElementsNeeded = 0;
	double* inOffsetsNeededPr = 0;
	if (nrhs > 1 && mxGetN(prhs[1]) == 2) {
		numElementsNeeded = mxGetM(prhs[1]);
		inOffsetsNeededPr = mxGetPr(prhs[1]);
	}
	long long* offsetsNeeded = new long long[numElementsNeeded];
	for (long long i = 0; i < numElementsNeeded; i++) {
		long long iRow = floor(inOffsetsNeededPr[i] + 0.5);
		long long iCol = floor(inOffsetsNeededPr[i + numElementsNeeded] + 0.5);
		offsetsNeeded[i] = (iCol - 1) * n + iRow;
	}
	long long* offsetsCov = new long long[longNnz * 10];
	long long numElementsCov = 0;
	// Execute the routine
	if ((nrhs == 1) ||
		((nrhs == 2) && (mxGetN(prhs[1]) == 2)) ||
		((nrhs == 2) && (mxGetNumberOfElements(prhs[1]) == 1) && (mxGetScalar(prhs[1]) > 0)) ||
		((nrhs == 3) && (mxGetScalar(prhs[2]) > 0))) {
		DetermineElements(lowerCholesky, offsetsNeeded, numElementsNeeded, offsetsCov, &numElementsCov);
	}
	else {
		DetermineElementsThrifty(lowerCholesky, offsetsNeeded, numElementsNeeded, offsetsCov, &numElementsCov);
	}
	long long longCovNnz = numElementsCov;
	int intCovN = n;
	double* covElements = new double[longCovNnz];
	for (long long i = 0; i < longCovNnz; i++) {
		covElements[i] = 1.;
	}
	ConstSparseSqMatrix lowerCov = ConstSparseSqMatrix(offsetsCov, covElements, &longCovNnz, &intCovN);
	// Execute the routine
	FillCov(lowerCholesky, lowerCov);
	// Prepare output data
	plhs[0] = mxCreateSparse(n, n, longCovNnz, mxREAL);
	double* outPr = mxGetPr(plhs[0]);
	mwIndex* outIr = mxGetIr(plhs[0]);
	mwIndex* outJc = mxGetJc(plhs[0]);
	long long prevCol = -1;
	for (long long i = 0; i < longCovNnz; i++) {
		outPr[i] = lowerCov.elementsByCol[i];
		outIr[i] = (offsetsCov[i] - 1) % n;
		long long curCol = (offsetsCov[i] - 1) / n;
		if (curCol > prevCol) {
			for (long long io = prevCol + 1; io <= curCol; io++) {
				outJc[io] = i;
			}
			prevCol = curCol;
		}
	}
	outJc[n] = longCovNnz;
	// Release memory
	delete[] inOffsets;
	delete[] offsetsNeeded;
	delete[] offsetsCov;
	delete[] covElements;
}
