#include "mex.h"
#include <vector>
#include <algorithm>
#include <math.h>

using namespace std;

/* -------------------------------------------------------------------------- */
/* Class-container for a sparse square matrix                                 */
/* -------------------------------------------------------------------------- */
class ConstSparseSqMatrix {
public:
	int numCol, numRow;
	long long numElements;
	double* elementsByCol;
	int* rowLength, *colLength, *colIndex, *rowIndex;
	long long* offsets, *colOffset;
	int** indexInCol, **indexInRow, **offsetInCol, **offsetInRow;
	ConstSparseSqMatrix(long long*, double*, long long*, int*);
	~ConstSparseSqMatrix() {
		delete[] this->colLength;
		delete[] this->colOffset;
		delete[] this->rowLength;
		delete[] this->colIndex;
		delete[] this->rowIndex;
		for (int i = 0; i < this->numCol; i++) {
			delete[] indexInCol[i];
			delete[] offsetInCol[i];
		}
		delete[] this->indexInCol;
		delete[] this->offsetInCol;
		for (int i = 0; i < this->numRow; i++) {
			delete[] indexInRow[i];
			delete[] offsetInRow[i];
		}
		delete[] this->indexInRow;
		delete[] this->offsetInRow;
	}
};

/* -------------------------------------------------------------------------- */
/* Template-container for a lower triangular indicator matrix.                */
/* Only indicators of (n + 1) * n / 2 cells are stored, each as a single bit  */
/* The class hs only three functions: constructor, set, and check             */
/* -------------------------------------------------------------------------- */
template<typename BlockType> class binaryMatrixLower {
private:
	binaryMatrixLower();
protected:
	int n, blockSize;
	BlockType *body;
	unsigned long long size;
public:
	binaryMatrixLower(int);
	bool setIfNotSet(int, int);
	bool isSet(int, int);
	void free() {
		if (this->body) {
			delete this->body;
			this->body = NULL;
		}
	}
	~binaryMatrixLower() {
		free();
	}
};

/* -------------------------------------------------------------------------- */
/* Implementation of template "binaryMatrixLower" (constructor)               */
/* -------------------------------------------------------------------------- */
template<typename BlockType> binaryMatrixLower<BlockType>::
binaryMatrixLower(int n) {
	this->n = n; this->blockSize = 8 * sizeof(BlockType);
	this->size = (((unsigned long long)n * ((unsigned long long)n + 1)) / 2) / blockSize + 1;
	body = new BlockType[this->size]();
}

/* -------------------------------------------------------------------------- */
/* Implementation of template "binaryMatrixLower" (set)                       */
/* -------------------------------------------------------------------------- */
template<typename BlockType> bool binaryMatrixLower<BlockType>::
setIfNotSet(int iRow, int iCol) {
	unsigned long long bitNumber = (unsigned long long)n * (unsigned long long)iCol + (unsigned long long)iRow - ((unsigned long long)iCol * ((unsigned long long)iCol + 1)) / 2;
	unsigned long long blockNumber = bitNumber / blockSize;
	int bitInBlock = bitNumber % blockSize;
	BlockType mask = (BlockType)1 << bitInBlock;
	if (body[blockNumber] & mask) {
		return false;
	}
	else {
		body[blockNumber] |= mask;
		return true;
	}
}

/* -------------------------------------------------------------------------- */
/* Implementation of template "binaryMatrixLower" (check)                     */
/* -------------------------------------------------------------------------- */
template<typename BlockType> bool binaryMatrixLower<BlockType>::
isSet(int iRow, int iCol) {
	unsigned long long bitNumber = (unsigned long long)n * (unsigned long long)iCol + (unsigned long long)iRow - ((unsigned long long)iCol * ((unsigned long long)iCol + 1)) / 2;
	unsigned long long blockNumber = bitNumber / blockSize;
	int bitInBlock = bitNumber % blockSize;
	BlockType mask = (BlockType)1 << bitInBlock;
	if (body[blockNumber] & mask) {
		return true;
	}
	else {
		return false;
	}
}

// Routines' declarations

/* -------------------------------------------------------------------------- */
/* Determine elements of the covariance matrix to be calculated.              */
/* The routine is fast due to the use of "binaryMatrixLower", but             */
/* requires ceil(n * (n + 1) / 2 / 8) bytes of physical memory                */
/* -------------------------------------------------------------------------- */
int DetermineElements(ConstSparseSqMatrix &m, long long* offsetsNeeded,
	long long numElementsNeeded, long long* offsets, long long* numOffsets);

/* -------------------------------------------------------------------------- */
/* Determine elements of the covariance matrix to be calculated.              */
/* The routine allocates memory dynamically for nonzero elements only, which  */
/* makes it slower than when using "binaryMatrixLower"                        */
/* -------------------------------------------------------------------------- */
int DetermineElementsThrifty(ConstSparseSqMatrix &m, long long* offsetsNeeded,
	long long numElementsNeeded, long long* offsets, long long* numOffsets);

/* -------------------------------------------------------------------------- */
/* Calculate all the nonzero entries of the matrix                            */
/* -------------------------------------------------------------------------- */
int FillCov(ConstSparseSqMatrix &lowerCholesky, ConstSparseSqMatrix &lowerCov);
